# -*- coding: utf-8 -*-
"""
Natsort can sort strings with numbers in a natural order.
It provides the natsorted function to sort strings with
arbitrary numbers.

You can mix types with natsorted.  This can get around the new
'unorderable types' issue with Python 3. Natsort will recursively
descend into lists of lists so you can sort by the sublist contents.

See the README or the natsort homepage for more details.

"""

from __future__ import (print_function, division,
                        unicode_literals, absolute_import)

import re
from os import curdir, pardir
from os.path import split, splitext
from operator import itemgetter
from functools import partial
from itertools import islice
from warnings import warn

from .py23compat import u_format, py23_str, py23_zip

# Make sure the doctest works for either python2 or python3
__doc__ = u_format(__doc__)

# The regex that locates floats
float_sign_exp_re = re.compile(r'([-+]?\d*\.?\d+(?:[eE][-+]?\d+)?)')
float_nosign_exp_re = re.compile(r'(\d*\.?\d+(?:[eE][-+]?\d+)?)')
float_sign_noexp_re = re.compile(r'([-+]?\d*\.?\d+)')
float_nosign_noexp_re = re.compile(r'(\d*\.?\d+)')
# Integer regexes
int_nosign_re = re.compile(r'(\d+)')
int_sign_re = re.compile(r'([-+]?\d+)')
# This dict will help select the correct regex and number conversion function.
regex_and_num_function_chooser = {
    (float, True,  True):  (float_sign_exp_re,     float),
    (float, True,  False): (float_sign_noexp_re,   float),
    (float, False, True):  (float_nosign_exp_re,   float),
    (float, False, False): (float_nosign_noexp_re, float),
    (int,   True,  True):  (int_sign_re,   int),
    (int,   True,  False): (int_sign_re,   int),
    (int,   False, True):  (int_nosign_re, int),
    (int,   False, False): (int_nosign_re, int),
    (None,  True,  True):  (int_nosign_re, int),
    (None,  True,  False): (int_nosign_re, int),
    (None,  False, True):  (int_nosign_re, int),
    (None,  False, False): (int_nosign_re, int),
}

# Number types.  I have to use set([...]) and not {...}
# because I am supporting Python 2.6.
number_types = set([float, int])

# This regex is to make sure we don't mistake a number for a file extension
decimal = re.compile(r'\.\d')


def _number_finder(s, regex, numconv, py3_safe):
    """Helper to split numbers"""

    # Split the input string by numbers.
    # If there are no splits, return now.
    # If the input is not a string, ValueError is raised.
    s = regex.split(s)
    if len(s) == 1:
        return tuple(s)

    # Now convert the numbers to numbers, and leave strings as strings.
    # Remove empty strings from the list.
    # Profiling showed that using regex here is much faster than
    # try/except with the numconv function.
    r = regex.match
    s = [numconv(x) if r(x) else x for x in s if x]

    # If the list begins with a number, lead with an empty string.
    # This is used to get around the "unorderable types" issue.
    # The most common case will be a string at the front of the
    # list, and in that case the try/except method is faster than
    # using isinstance. This was chosen at the expense of the less
    # common case of a number being at the front of the list.
    try:
        s[0][0]  # str supports indexing, but not numbers
    except TypeError:
        s = [''] + s

    # The _py3_safe function inserts "" between numbers in the list,
    # and is used to get around "unorderable types" in complex cases.
    # It is a separate function that needs to be requested specifically
    # because it is expensive to call.
    return _py3_safe(s) if py3_safe else s


def _path_splitter(s):
    """Split a string into its path components. Assumes a string is a path."""
    path_parts = []
    p_append = path_parts.append
    path_location = s
    # Continue splitting the path from the back until we have reached
    # '..' or '.', or until there is nothing left to split.
    while path_location != curdir and path_location != pardir:
        parent_path = path_location
        path_location, child_path = split(parent_path)
        if path_location == parent_path:
            break
        p_append(child_path)
    # This last append is the base path.
    # Only append if the string is non-empty.
    if path_location:
        p_append(path_location)
    # We created this list in reversed order, so we now correct the order.
    path_parts.reverse()
    # Now, split off the file extensions using a similar method to above.
    # Continue splitting off file extensions until we reach a decimal number
    # or there are no more extensions.
    base = path_parts.pop()
    base_parts = []
    b_append = base_parts.append
    d_match = decimal.match
    while True:
        front = base
        base, ext = splitext(front)
        if d_match(ext) or not ext:
            # Reset base to before the split if the split is invalid.
            base = front
            break
        b_append(ext)
    b_append(base)
    base_parts.reverse()
    # Return the split parent paths and then the split basename.
    return path_parts + base_parts


def _py3_safe(parsed_list):
    """Insert '' between two numbers."""
    length = len(parsed_list)
    if length < 2:
        return parsed_list
    else:
        new_list = [parsed_list[0]]
        nl_append = new_list.append
        ntypes = number_types
        for before, after in py23_zip(islice(parsed_list, 0, length-1),
                                      islice(parsed_list, 1, None)):
            # I realize that isinstance is favored over type, but
            # in this case type is SO MUCH FASTER than isinstance!!
            if type(before) in ntypes and type(after) in ntypes:
                nl_append("")
            nl_append(after)
        return new_list


def _natsort_key(val, key=None, number_type=float, signed=True, exp=True,
                 as_path=False, py3_safe=False):
    """\
    Key to sort strings and numbers naturally.

    It works by separating out the numbers from the strings. This function for
    internal use only. See the natsort_keygen documentation for details of each
    parameter.

    Parameters
    ----------
    val : {str, unicode}
    key : callable, optional
    number_type : {None, float, int}, optional
    signed : {True, False}, optional
    exp : {True, False}, optional
    as_path : {True, False}, optional
    py3_safe : {True, False}, optional

    Returns
    -------
    out : tuple
        The modified value with numbers extracted.

    """

    # Convert the arguments to the proper input tuple
    inp_options = (number_type, signed, exp)
    try:
        regex, num_function = regex_and_num_function_chooser[inp_options]
    except KeyError:
        # Report errors properly
        if number_type not in (float, int) and number_type is not None:
            raise ValueError("_natsort_key: 'number_type' parameter "
                             "'{0}' invalid".format(py23_str(number_type)))
        elif signed not in (True, False):
            raise ValueError("_natsort_key: 'signed' parameter "
                             "'{0}' invalid".format(py23_str(signed)))
        elif exp not in (True, False):
            raise ValueError("_natsort_key: 'exp' parameter "
                             "'{0}' invalid".format(py23_str(exp)))
    else:
        # Apply key if needed.
        if key is not None:
            val = key(val)

        # If this is a path, convert it.
        # An AttrubuteError is raised if not a string.
        split_as_path = False
        if as_path:
            try:
                val = _path_splitter(val)
            except AttributeError:
                pass
            else:
                # Record that this string was split as a path so that
                # we can set as_path to False in the recursive call.
                split_as_path = True

        # Assume the input are strings, which is the most common case.
        try:
            return tuple(_number_finder(val, regex, num_function, py3_safe))
        except TypeError:
            # If not strings, assume it is an iterable that must
            # be parsed recursively. Do not apply the key recursively.
            # If this string was split as a path, set as_path to False.
            try:
                return tuple([_natsort_key(x, None, number_type, signed,
                                           exp, as_path and not split_as_path,
                                           py3_safe) for x in val])
            # If there is still an error, it must be a number.
            # Return as-is, with a leading empty string.
            # Waiting for two raised errors instead of calling
            # isinstance at the opening of the function is slower
            # for numbers but much faster for strings, and since
            # numbers are not a common input to natsort this is
            # an acceptable sacrifice.
            except TypeError:
                return (('', val,),) if as_path else ('', val,)


@u_format
def natsort_key(val, key=None, number_type=float, signed=True, exp=True,
                as_path=False, py3_safe=False):
    """\
    Key to sort strings and numbers naturally.

    Key to sort strings and numbers naturally, not lexicographically.
    It is designed for use in passing to the 'sorted' builtin or
    'sort' attribute of lists.

    .. note:: Depreciated since version 3.4.0.
              This function remains in the publicly exposed API for
              backwards-compatibility reasons, but future development
              should use the newer `natsort_keygen` function. It is
              planned to remove this from the public API in natsort
              version 4.0.0.  A DeprecationWarning will be raised
              via the warnings module; set warnings.simplefilter("always")
              to raise them to see if your code will work in version
              4.0.0.

    Parameters
    ----------
    val : {{str, unicode}}
        The value used by the sorting algorithm

    key : callable, optional
        A key used to manipulate the input value before parsing for
        numbers. It is **not** applied recursively.
        It should accept a single argument and return a single value.

    number_type : {{None, float, int}}, optional
        The types of number to sort on: `float` searches for floating
        point numbers, `int` searches for integers, and `None` searches
        for digits (like integers but does not take into account
        negative sign). `None` is a shortcut for `number_type = int`
        and `signed = False`.

    signed : {{True, False}}, optional
        By default a '+' or '-' before a number is taken to be the sign
        of the number. If `signed` is `False`, any '+' or '-' will not
        be considered to be part of the number, but as part part of the
        string.

    exp : {{True, False}}, optional
        This option only applies to `number_type = float`.  If
        `exp = True`, a string like "3.5e5" will be interpreted as
        350000, i.e. the exponential part is considered to be part of
        the number. If `exp = False`, "3.5e5" is interpreted as
        ``(3.5, "e", 5)``. The default behavior is `exp = True`.

    as_path : {{True, False}}, optional
        This option will force strings to be interpreted as filesystem
        paths, so they will be split according to the filesystem separator
        (i.e. '/' on UNIX, '\\\\' on Windows), as well as splitting on the
        file extension, if any. Without this, lists of file paths like
        ``['Folder', 'Folder (1)', 'Folder (10)']`` will not be sorted
        properly; ``'Folder'`` will be placed at the end, not at the front.
        The default behavior is `as_path = False`.

    py3_safe : {{True, False}}, optional
        This will make the string parsing algorithm be more careful by
        placing an empty string between two adjacent numbers after the
        parsing algorithm. This will prevent the "unorderable types"
        error.

    Returns
    -------
    out : tuple
        The modified value with numbers extracted.

    See Also
    --------
    natsort_keygen : Generates a properly wrapped `natsort_key`.

    Examples
    --------
    Using natsort_key is just like any other sorting key in python::

        >>> a = ['num3', 'num5', 'num2']
        >>> a.sort(key=natsort_key)
        >>> a
        [{u}'num2', {u}'num3', {u}'num5']

    It works by separating out the numbers from the strings::

        >>> natsort_key('num2')
        ({u}'num', 2.0)

    If you need to call natsort_key with the number_type argument, or get a
    special attribute or item of each element of the sequence, please use
    the `natsort_keygen` function.  Actually, please just use the
    `natsort_keygen` function.

    Notes
    -----
    Iterables are parsed recursively so you can sort lists of lists::

        >>> natsort_key(('a1', 'a10'))
        (({u}'a', 1.0), ({u}'a', 10.0))

    Strings that lead with a number get an empty string at the front of the
    tuple. This is designed to get around the "unorderable types" issue of
    Python3::

        >>> natsort_key('15a')
        ({u}'', 15.0, {u}'a')

    You can give bare numbers, too::

        >>> natsort_key(10)
        ({u}'', 10)

    If you have a case where one of your string has two numbers in a row,
    you can turn on the "py3_safe" option to try to add a "" between sets
    of two numbers::

        >>> natsort_key('43h7+3', py3_safe=True)
        ({u}'', 43.0, {u}'h', 7.0, {u}'', 3.0)

    """
    msg = "natsort_key is depreciated as of 3.4.0, please use natsort_keygen"
    warn(msg, DeprecationWarning)
    return _natsort_key(val, key, number_type, signed, exp, as_path, py3_safe)


@u_format
def natsort_keygen(key=None, number_type=float, signed=True, exp=True,
                   as_path=False, py3_safe=False):
    """\
    Generate a key to sort strings and numbers naturally.

    Generate a key to sort strings and numbers naturally,
    not lexicographically. This key is designed for use as the
    `key` argument to functions such as the `sorted` builtin.

    The user may customize the generated function with the
    arguments to `natsort_keygen`, including an optional
    `key` function which will be called before the `natsort_key`.

    Parameters
    ----------
    key : callable, optional
        A key used to manipulate the input value before parsing for
        numbers. It is **not** applied recursively.
        It should accept a single argument and return a single value.

    number_type : {{None, float, int}}, optional
        The types of number to sort on: `float` searches for floating
        point numbers, `int` searches for integers, and `None` searches
        for digits (like integers but does not take into account
        negative sign). `None` is a shortcut for `number_type = int`
        and `signed = False`.

    signed : {{True, False}}, optional
        By default a '+' or '-' before a number is taken to be the sign
        of the number. If `signed` is `False`, any '+' or '-' will not
        be considered to be part of the number, but as part part of the
        string.

    exp : {{True, False}}, optional
        This option only applies to `number_type = float`.  If
        `exp = True`, a string like "3.5e5" will be interpreted as
        350000, i.e. the exponential part is considered to be part of
        the number. If `exp = False`, "3.5e5" is interpreted as
        ``(3.5, "e", 5)``. The default behavior is `exp = True`.

    as_path : {{True, False}}, optional
        This option will force strings to be interpreted as filesystem
        paths, so they will be split according to the filesystem separator
        (i.e. `/` on UNIX, `\\\\` on Windows), as well as splitting on the
        file extension, if any. Without this, lists with file paths like
        ``['Folder/', 'Folder (1)/', 'Folder (10)/']`` will not be sorted
        properly; ``'Folder'`` will be placed at the end, not at the front.
        The default behavior is `as_path = False`.

    py3_safe : {{True, False}}, optional
        This will make the string parsing algorithm be more careful by
        placing an empty string between two adjacent numbers after the
        parsing algorithm. This will prevent the "unorderable types"
        error.

    Returns
    -------
    out : function
        A wrapped version of the `natsort_key` function that is
        suitable for passing as the `key` argument to functions
        such as `sorted`.

    Examples
    --------
    `natsort_keygen` is a convenient waynto create a custom key
    to sort lists in-place (for example). Calling with no objects
    will return a plain `natsort_key` instance::

        >>> a = ['num5.10', 'num-3', 'num5.3', 'num2']
        >>> b = a[:]
        >>> a.sort(key=natsort_key)
        >>> b.sort(key=natsort_keygen())
        >>> a == b
        True

    The power of `natsort_keygen` is when you want to want to pass
    arguments to the `natsort_key`.  Consider the following
    equivalent examples; which is more clear? ::

        >>> a = ['num5.10', 'num-3', 'num5.3', 'num2']
        >>> b = a[:]
        >>> a.sort(key=lambda x: natsort_key(x, key=lambda y: y.upper(),
        ...        signed=False))
        >>> b.sort(key=natsort_keygen(key=lambda x: x.upper(), signed=False))
        >>> a == b
        True

    """
    return partial(_natsort_key,
                   key=key,
                   number_type=number_type,
                   signed=signed,
                   exp=exp,
                   as_path=as_path,
                   py3_safe=py3_safe)


@u_format
def natsorted(seq, key=None, number_type=float, signed=True, exp=True,
              reverse=False, as_path=False):
    """\
    Sorts a sequence naturally.

    Sorts a sequence naturally (alphabetically and numerically),
    not lexicographically. Returns a new copy of the sorted
    sequence as a list.

    Parameters
    ----------
    seq : iterable
        The sequence to sort.

    key : callable, optional
        A key used to determine how to sort each element of the sequence.
        It is **not** applied recursively.
        It should accept a single argument and return a single value.

    number_type : {{None, float, int}}, optional
        The types of number to sort on: `float` searches for floating
        point numbers, `int` searches for integers, and `None` searches
        for digits (like integers but does not take into account
        negative sign). `None` is a shortcut for `number_type = int`
        and `signed = False`.

    signed : {{True, False}}, optional
        By default a '+' or '-' before a number is taken to be the sign
        of the number. If `signed` is `False`, any '+' or '-' will not
        be considered to be part of the number, but as part part of the
        string.

    exp : {{True, False}}, optional
        This option only applies to `number_type = float`.  If
        `exp = True`, a string like "3.5e5" will be interpreted as
        350000, i.e. the exponential part is considered to be part of
        the number. If `exp = False`, "3.5e5" is interpreted as
        ``(3.5, "e", 5)``. The default behavior is `exp = True`.

    reverse : {{True, False}}, optional
        Return the list in reversed sorted order. The default is
        `False`.

    as_path : {{True, False}}, optional
        This option will force strings to be interpreted as filesystem
        paths, so they will be split according to the filesystem separator
        (i.e. '/' on UNIX, '\\\\' on Windows), as well as splitting on the
        file extension, if any. Without this, lists of file paths like
        ``['Folder', 'Folder (1)', 'Folder (10)']`` will not be sorted
        properly; ``'Folder'`` will be placed at the end, not at the front.
        The default behavior is `as_path = False`.

    Returns
    -------
    out: list
        The sorted sequence.

    See Also
    --------
    natsort_keygen : Generates the key that makes natural sorting possible.
    versorted : A wrapper for ``natsorted(seq, number_type=None)``.
    index_natsorted : Returns the sorted indexes from `natsorted`.

    Examples
    --------
    Use `natsorted` just like the builtin `sorted`::

        >>> a = ['num3', 'num5', 'num2']
        >>> natsorted(a)
        [{u}'num2', {u}'num3', {u}'num5']

    """
    try:
        return sorted(seq, reverse=reverse,
                      key=natsort_keygen(key, number_type,
                                         signed, exp, as_path))
    except TypeError as e:
        # In the event of an unresolved "unorderable types" error
        # attempt to sort again, being careful to prevent this error.
        if 'unorderable types' in str(e):
            return sorted(seq, reverse=reverse,
                          key=natsort_keygen(key, number_type,
                                             signed, exp, as_path,
                                             True))
        else:
            # Re-raise if the problem was not "unorderable types"
            raise


@u_format
def versorted(seq, key=None, reverse=False, as_path=False):
    """\
    Convenience function to sort version numbers.

    Convenience function to sort version numbers. This is a wrapper
    around ``natsorted(seq, number_type=None)``.

    Parameters
    ----------
    seq : iterable
        The sequence to sort.

    key : callable, optional
        A key used to determine how to sort each element of the sequence.
        It is **not** applied recursively.
        It should accept a single argument and return a single value.

    reverse : {{True, False}}, optional
        Return the list in reversed sorted order. The default is
        `False`.

    as_path : {{True, False}}, optional
        This option will force strings to be interpreted as filesystem
        paths, so they will be split according to the filesystem separator
        (i.e. '/' on UNIX, '\\\\' on Windows), as well as splitting on the
        file extension, if any. Without this, lists of file paths like
        ``['Folder', 'Folder (1)', 'Folder (10)']`` will not be sorted
        properly; ``'Folder'`` will be placed at the end, not at the front.
        The default behavior is `as_path = False`.

    Returns
    -------
    out : list
        The sorted sequence.

    See Also
    --------
    index_versorted : Returns the sorted indexes from `versorted`.

    Examples
    --------
    Use `versorted` just like the builtin `sorted`::

        >>> a = ['num4.0.2', 'num3.4.1', 'num3.4.2']
        >>> versorted(a)
        [{u}'num3.4.1', {u}'num3.4.2', {u}'num4.0.2']

    """
    return natsorted(seq, key, None, reverse=reverse, as_path=as_path)


@u_format
def index_natsorted(seq, key=None, number_type=float, signed=True, exp=True,
                    reverse=False, as_path=False):
    """\
    Return the list of the indexes used to sort the input sequence.

    Sorts a sequence naturally, but returns a list of sorted the
    indexes and not the sorted list. This list of indexes can be
    used to sort multiple lists by the sorted order of the given
    sequence.

    Parameters
    ----------
    seq : iterable
        The sequence to sort.

    key : callable, optional
        A key used to determine how to sort each element of the sequence.
        It is **not** applied recursively.
        It should accept a single argument and return a single value.

    number_type : {{None, float, int}}, optional
        The types of number to sort on: `float` searches for floating
        point numbers, `int` searches for integers, and `None` searches
        for digits (like integers but does not take into account
        negative sign). `None` is a shortcut for `number_type = int`
        and `signed = False`.

    signed : {{True, False}}, optional
        By default a '+' or '-' before a number is taken to be the sign
        of the number. If `signed` is `False`, any '+' or '-' will not
        be considered to be part of the number, but as part part of the
        string.

    exp : {{True, False}}, optional
        This option only applies to `number_type = float`.  If
        `exp = True`, a string like "3.5e5" will be interpreted as
        350000, i.e. the exponential part is considered to be part of
        the number. If `exp = False`, "3.5e5" is interpreted as
        ``(3.5, "e", 5)``. The default behavior is `exp = True`.

    reverse : {{True, False}}, optional
        Return the list in reversed sorted order. The default is
        `False`.

    as_path : {{True, False}}, optional
        This option will force strings to be interpreted as filesystem
        paths, so they will be split according to the filesystem separator
        (i.e. '/' on UNIX, '\\\\' on Windows), as well as splitting on the
        file extension, if any. Without this, lists of file paths like
        ``['Folder', 'Folder (1)', 'Folder (10)']`` will not be sorted
        properly; ``'Folder'`` will be placed at the end, not at the front.
        The default behavior is `as_path = False`.

    Returns
    -------
    out : tuple
        The ordered indexes of the sequence.

    See Also
    --------
    natsorted
    order_by_index

    Examples
    --------

    Use index_natsorted if you want to sort multiple lists by the
    sorted order of one list::

        >>> a = ['num3', 'num5', 'num2']
        >>> b = ['foo', 'bar', 'baz']
        >>> index = index_natsorted(a)
        >>> index
        [2, 0, 1]
        >>> # Sort both lists by the sort order of a
        >>> order_by_index(a, index)
        [{u}'num2', {u}'num3', {u}'num5']
        >>> order_by_index(b, index)
        [{u}'baz', {u}'foo', {u}'bar']

    """
    if key is None:
        newkey = itemgetter(1)
    else:
        newkey = lambda x: key(itemgetter(1)(x))
    # Pair the index and sequence together, then sort by element
    index_seq_pair = [[x, y] for x, y in enumerate(seq)]
    try:
        index_seq_pair.sort(reverse=reverse,
                            key=natsort_keygen(newkey, number_type,
                                               signed, exp, as_path))
    except TypeError as e:
        # In the event of an unresolved "unorderable types" error
        # attempt to sort again, being careful to prevent this error.
        if 'unorderable types' in str(e):
            index_seq_pair.sort(reverse=reverse,
                                key=natsort_keygen(newkey, number_type,
                                                   signed, exp, as_path,
                                                   True))
        else:
            # Re-raise if the problem was not "unorderable types"
            raise
    return [x for x, _ in index_seq_pair]


@u_format
def index_versorted(seq, key=None, reverse=False, as_path=False):
    """\
    Return the list of the indexes used to sort the input sequence
    of version numbers.

    Sorts a sequence naturally, but returns a list of sorted the
    indexes and not the sorted list. This list of indexes can be
    used to sort multiple lists by the sorted order of the given
    sequence.

    This is a wrapper around ``index_natsorted(seq, number_type=None)``.

    Parameters
    ----------
    seq: iterable
        The sequence to sort.

    key: callable, optional
        A key used to determine how to sort each element of the sequence.
        It is **not** applied recursively.
        It should accept a single argument and return a single value.

    reverse : {{True, False}}, optional
        Return the list in reversed sorted order. The default is
        `False`.

    as_path : {{True, False}}, optional
        This option will force strings to be interpreted as filesystem
        paths, so they will be split according to the filesystem separator
        (i.e. '/' on UNIX, '\\\\' on Windows), as well as splitting on the
        file extension, if any. Without this, lists of file paths like
        ``['Folder', 'Folder (1)', 'Folder (10)']`` will not be sorted
        properly; ``'Folder'`` will be placed at the end, not at the front.
        The default behavior is `as_path = False`.

    Returns
    -------
    out : tuple
        The ordered indexes of the sequence.

    See Also
    --------
    versorted
    order_by_index

    Examples
    --------
    Use `index_versorted` just like the builtin `sorted`::

        >>> a = ['num4.0.2', 'num3.4.1', 'num3.4.2']
        >>> index_versorted(a)
        [1, 2, 0]

    """
    return index_natsorted(seq, key, None, reverse=reverse, as_path=as_path)


@u_format
def order_by_index(seq, index, iter=False):
    """\
    Order a given sequence by an index sequence.

    The output of `index_natsorted` and `index_versorted` is a
    sequence of integers (index) that correspond to how its input
    sequence **would** be sorted. The idea is that this index can
    be used to reorder multiple sequences by the sorted order of the
    first sequence. This function is a convenient wrapper to
    apply this ordering to a sequence.

    Parameters
    ----------
    seq : iterable
        The sequence to order.

    index : iterable
        The sequence that indicates how to order `seq`.
        It should be the same length as `seq` and consist
        of integers only.

    iter : {{True, False}}, optional
        If `True`, the ordered sequence is returned as a
        generator expression; otherwise it is returned as a
        list. The default is `False`.

    Returns
    -------
    out : {{list, generator}}
        The sequence ordered by `index`, as a `list` or as a
        generator expression (depending on the value of `iter`).

    See Also
    --------
    index_natsorted
    index_versorted

    Examples
    --------

    `order_by_index` is a comvenience function that helps you apply
    the result of `index_natsorted` or `index_versorted`::

        >>> a = ['num3', 'num5', 'num2']
        >>> b = ['foo', 'bar', 'baz']
        >>> index = index_natsorted(a)
        >>> index
        [2, 0, 1]
        >>> # Sort both lists by the sort order of a
        >>> order_by_index(a, index)
        [{u}'num2', {u}'num3', {u}'num5']
        >>> order_by_index(b, index)
        [{u}'baz', {u}'foo', {u}'bar']

    """
    return (seq[i] for i in index) if iter else [seq[i] for i in index]



# -*- coding: utf-8 -*-
from __future__ import (print_function, division,
                        unicode_literals, absolute_import)

import sys

from .natsort import natsorted, regex_and_num_function_chooser
from ._version import __version__
from .py23compat import py23_str


def main():
    """\
    Performs a natural sort on entries given on the command-line.
    A natural sort sorts numerically then alphabetically, and will sort
    by numbers in the middle of an entry.
    """

    from argparse import ArgumentParser, RawDescriptionHelpFormatter
    from textwrap import dedent
    parser = ArgumentParser(description=dedent(main.__doc__),
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('--version', action='version',
                        version='%(prog)s {0}'.format(__version__))
    parser.add_argument(
        '-p', '--paths', default=False, action='store_true',
        help='Interpret the input as file paths.  This is not '
             'strictly necessary to sort all file paths, but in cases '
             'where there are OS-generated file paths like "Folder/" '
             'and "Folder (1)/", this option is needed to make the '
             'paths sorted in the order you expect ("Folder/" before '
             '"Folder (1)/").')
    parser.add_argument(
        '-f', '--filter', nargs=2, type=float, metavar=('LOW', 'HIGH'),
        action='append',
        help='Used for keeping only the entries that have a number '
             'falling in the given range.')
    parser.add_argument(
        '-F', '--reverse-filter', nargs=2, type=float,
        metavar=('LOW', 'HIGH'), action='append', dest='reverse_filter',
        help='Used for excluding the entries that have a number '
             'falling in the given range.')
    parser.add_argument(
        '-e', '--exclude', type=float, action='append',
        help='Used to exclude an entry that contains a specific number.')
    parser.add_argument(
        '-r', '--reverse', action='store_true', default=False,
        help='Returns in reversed order.')
    parser.add_argument(
        '-t', '--number-type', '--number_type', dest='number_type',
        choices=('digit', 'int', 'float', 'version', 'ver'), default='float',
        help='Choose the type of number to search for. "float" will search '
             'for floating-point numbers.  "int" will only search for '
             'integers. "digit", "version", and "ver" are shortcuts for "int" '
             'with --nosign.')
    parser.add_argument(
        '--nosign', default=True, action='store_false', dest='signed',
        help='Do not consider "+" or "-" as part of a number, i.e. do not '
             'take sign into consideration.')
    parser.add_argument(
        '--noexp', default=True, action='store_false', dest='exp',
        help='Do not consider an exponential as part of a number, i.e. 1e4, '
             'would be considered as 1, "e", and 4, not as 10000.  This only '
             'effects the --number-type=float.')
    parser.add_argument(
        'entries', nargs='*', default=sys.stdin,
        help='The entries to sort. Taken from stdin if nothing is given on '
             'the command line.', )
    args = parser.parse_args()

    # Make sure the filter range is given properly. Does nothing if no filter
    args.filter = check_filter(args.filter)

    # Remove trailing whitespace from all the entries
    entries = [e.strip() for e in args.entries]

    # Sort by directory then by file within directory and print.
    sort_and_print_entries(entries, args)


def range_check(low, high):
    """\
    Verifies that that given range has a low lower than the high.
    If the condition is not met, a ValueError is raised.
    Otherwise, the values are returned, but as floats.
    """
    low, high = float(low), float(high)
    if low >= high:
        raise ValueError('low >= high')
    else:
        return low, high


def check_filter(filt):
    """\
    Check that the low value of the filter is lower than the high.
    If there is to be no filter, return 'None'.
    If the condition is not met, a ValueError is raised.
    Otherwise, the values are returned, but as floats.
    """
    # Quick return if no filter.
    if not filt:
        return None
    try:
        return [range_check(f[0], f[1]) for f in filt]
    except ValueError as a:
        raise ValueError('Error in --filter: '+py23_str(a))


def keep_entry_range(entry, lows, highs, converter, regex):
    """\
    Boolean function to determine if an entry should be kept out
    based on if any numbers are in a given range.

    Returns True if it should be kept (i.e. falls in the range),
    and False if it is not in the range and should not be kept.
    """
    return any(low <= converter(num) <= high
               for num in regex.findall(entry)
               for low, high in zip(lows, highs))


def exclude_entry(entry, values, converter, regex):
    """\
    Boolean function to determine if an entry should be kept out
    based on if it contains a specific number.

    Returns True if it should be kept (i.e. does not match),
    and False if it matches and should not be kept.
    """
    return not any(converter(num) in values for num in regex.findall(entry))


def sort_and_print_entries(entries, args):
    """Sort the entries, applying the filters first if necessary."""

    # Extract the proper number type.
    kwargs = {'number_type': {'digit': None,
                              'version': None,
                              'ver': None,
                              'int': int,
                              'float': float}[args.number_type],
              'signed': args.signed,
              'exp': args.exp,
              'as_path': args.paths,
              'reverse': args.reverse, }

    # Pre-remove entries that don't pass the filtering criteria
    # Make sure we use the same searching algorithm for filtering
    # as for sorting.
    do_filter = args.filter is not None or args.reverse_filter is not None
    if do_filter or args.exclude:
        inp_options = (kwargs['number_type'], args.signed, args.exp)
        regex, num_function = regex_and_num_function_chooser[inp_options]
        if args.filter is not None:
            lows, highs = ([f[0] for f in args.filter],
                           [f[1] for f in args.filter])
            entries = [entry for entry in entries
                       if keep_entry_range(entry, lows, highs,
                                           num_function, regex)]
        if args.reverse_filter is not None:
            lows, highs = ([f[0] for f in args.reverse_filter],
                           [f[1] for f in args.reverse_filter])
            entries = [entry for entry in entries
                       if not keep_entry_range(entry, lows, highs,
                                               num_function, regex)]
        if args.exclude:
            exclude = set(args.exclude)
            entries = [entry for entry in entries
                       if exclude_entry(entry, exclude,
                                        num_function, regex)]

    # Print off the sorted results
    for entry in natsorted(entries, **kwargs):
        print(entry)


if __name__ == '__main__':
    try:
        main()
    except ValueError as a:
        sys.exit(py23_str(a))
    except KeyboardInterrupt:
        sys.exit(1)



# -*- coding: utf-8 -*-
from __future__ import (print_function, division,
                        unicode_literals, absolute_import)

from .natsort import (natsort_key, natsort_keygen, natsorted,
                      index_natsorted, versorted, index_versorted,
                      order_by_index)
from ._version import __version__

__all__ = [
    'natsort_key',
    'natsort_keygen',
    'natsorted',
    'versorted'
    'index_natsorted',
    'index_versorted',
    'order_by_index',
]



# -*- coding: utf-8 -*-
"""\
Test the natsort command-line tool functions.
"""
import re
import sys
from pytest import raises
from natsort.__main__ import main, range_check, check_filter
from natsort.__main__ import keep_entry_range, exclude_entry
from natsort.__main__ import sort_and_print_entries


def test_main(capsys):

    # Simple sorting
    sys.argv[1:] = ['num-2', 'num-6', 'num-1']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
num-6
num-2
num-1
"""

    # Reverse order
    sys.argv[1:] = ['-r', 'num-2', 'num-6', 'num-1']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
num-1
num-2
num-6
"""

    # Neglect '-' or '+'
    sys.argv[1:] = ['--nosign', 'num-2', 'num-6', 'num-1']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
num-1
num-2
num-6
"""

    # Sort as digits
    sys.argv[1:] = ['-t', 'digit', 'num-2', 'num-6', 'num-1']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
num-1
num-2
num-6
"""

    # Sort as versions (synonym for digits)
    sys.argv[1:] = ['-t', 'version', 'num-2', 'num-6', 'num-1']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
num-1
num-2
num-6
"""

    # Exclude the number -1 and 6.  Only -1 is present.
    sys.argv[1:] = ['-t', 'int', '-e', '-1', '-e', '6',
                    'num-2', 'num-6', 'num-1']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
num-6
num-2
"""

    # Exclude the number 1 and 6.
    # Both are present because we use digits/versions.
    sys.argv[1:] = ['-t', 'ver', '-e', '1', '-e', '6',
                    'num-2', 'num-6', 'num-1']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
num-2
"""

    # Floats work too.
    sys.argv[1:] = ['a1.0e3', 'a5.3', 'a453.6']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
a5.3
a453.6
a1.0e3
"""

    # Only include in the range of 1-10.
    sys.argv[1:] = ['-f', '1', '10', 'a1.0e3', 'a5.3', 'a453.6']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
a5.3
"""

    # Don't include in the range of 1-10.
    sys.argv[1:] = ['-F', '1', '10', 'a1.0e3', 'a5.3', 'a453.6']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
a453.6
a1.0e3
"""

    # Include two ranges.
    sys.argv[1:] = ['-f', '1', '10', '-f', '400', '500',
                    'a1.0e3', 'a5.3', 'a453.6']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
a5.3
a453.6
"""

    # Don't account for exponential notation.
    sys.argv[1:] = ['--noexp', 'a1.0e3', 'a5.3', 'a453.6']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
a1.0e3
a5.3
a453.6
"""

    # To sort complicated filenames you need --paths
    sys.argv[1:] = ['/Folder (1)/', '/Folder/', '/Folder (10)/']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
/Folder (1)/
/Folder (10)/
/Folder/
"""
    sys.argv[1:] = ['--paths', '/Folder (1)/', '/Folder/', '/Folder (10)/']
    main()
    out, __ = capsys.readouterr()
    assert out == """\
/Folder/
/Folder (1)/
/Folder (10)/
"""


def test_range_check():

    # Floats are always returned
    assert range_check(10, 11) == (10.0, 11.0)
    assert range_check(6.4, 30) == (6.4, 30.0)

    # Invalid ranges give a ValueErro
    with raises(ValueError) as err:
        range_check(7, 2)
    assert str(err.value) == 'low >= high'


def test_check_filter():

    # No filter gives 'None'
    assert check_filter(()) is None
    assert check_filter(False) is None
    assert check_filter(None) is None

    # The check filter always returns floats
    assert check_filter([(6, 7)]) == [(6.0, 7.0)]
    assert check_filter([(6, 7), (2, 8)]) == [(6.0, 7.0), (2.0, 8.0)]

    # Invalid ranges get a ValueError
    with raises(ValueError) as err:
        check_filter([(7, 2)])
    assert str(err.value) == 'Error in --filter: low >= high'


def test_keep_entry_range():

    regex = re.compile(r'\d+')
    assert keep_entry_range('a56b23c89', [0], [100], int, regex)
    assert keep_entry_range('a56b23c89', [1, 88], [20, 90], int, regex)
    assert not keep_entry_range('a56b23c89', [1], [20], int, regex)


def test_exclude_entry():

    # Check if the exclude value is present in the input string
    regex = re.compile(r'\d+')
    assert exclude_entry('a56b23c89', [100, 45], int, regex)
    assert not exclude_entry('a56b23c89', [23], int, regex)


def test_sort_and_print_entries(capsys):

    class Args:
        """A dummy class to simulate the argparse Namespace object"""
        def __init__(self, filter, reverse_filter, exclude, as_path, reverse):
            self.filter = filter
            self.reverse_filter = reverse_filter
            self.exclude = exclude
            self.reverse = reverse
            self.number_type = 'float'
            self.signed = True
            self.exp = True
            self.paths = as_path

    entries = ['tmp/a57/path2',
               'tmp/a23/path1',
               'tmp/a1/path1',
               'tmp/a1 (1)/path1',
               'tmp/a130/path1',
               'tmp/a64/path1',
               'tmp/a64/path2']

    # Just sort the paths
    sort_and_print_entries(entries, Args(None, None, False, False, False))
    out, __ = capsys.readouterr()
    assert out == """\
tmp/a1 (1)/path1
tmp/a1/path1
tmp/a23/path1
tmp/a57/path2
tmp/a64/path1
tmp/a64/path2
tmp/a130/path1
"""

    # You would use --paths to make them sort
    # as paths when the OS makes duplicates
    sort_and_print_entries(entries, Args(None, None, False, True, False))
    out, __ = capsys.readouterr()
    assert out == """\
tmp/a1/path1
tmp/a1 (1)/path1
tmp/a23/path1
tmp/a57/path2
tmp/a64/path1
tmp/a64/path2
tmp/a130/path1
"""

    # Sort the paths with numbers between 20-100
    sort_and_print_entries(entries, Args([(20, 100)], None, False,
                                         False, False))
    out, __ = capsys.readouterr()
    assert out == """\
tmp/a23/path1
tmp/a57/path2
tmp/a64/path1
tmp/a64/path2
"""

    # Sort the paths without numbers between 20-100
    sort_and_print_entries(entries, Args(None, [(20, 100)], False,
                                         True, False))
    out, __ = capsys.readouterr()
    assert out == """\
tmp/a1/path1
tmp/a1 (1)/path1
tmp/a130/path1
"""

    # Sort the paths, excluding 23 and 130
    sort_and_print_entries(entries, Args(None, None, [23, 130], True, False))
    out, __ = capsys.readouterr()
    assert out == """\
tmp/a1/path1
tmp/a1 (1)/path1
tmp/a57/path2
tmp/a64/path1
tmp/a64/path2
"""

    # Sort the paths, excluding 2
    sort_and_print_entries(entries, Args(None, None, [2], False, False))
    out, __ = capsys.readouterr()
    assert out == """\
tmp/a1 (1)/path1
tmp/a1/path1
tmp/a23/path1
tmp/a64/path1
tmp/a130/path1
"""

    # Sort in reverse order
    sort_and_print_entries(entries, Args(None, None, False, True, True))
    out, __ = capsys.readouterr()
    assert out == """\
tmp/a130/path1
tmp/a64/path2
tmp/a64/path1
tmp/a57/path2
tmp/a23/path1
tmp/a1 (1)/path1
tmp/a1/path1
"""
